# -----------------------------------------------------------------------------
# Copyright 2019 Dave Jarvis
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Translates a number from digits to words using Chicago Manual of Style.
# This does not translate numbers greater than one hundred. If ordinal
# is TRUE, this will return the ordinal name. This will not produce ordinals
# for numbers greater than 100.
#
# @param n An integer value to potentially convert to English text
# @param ordinal Set TRUE to use ordinal numbers
# -----------------------------------------------------------------------------
cardinal <- function( n, ordinal = FALSE ) {
  if( n == 0 ) {
    if( ordinal ) {
      return( "zeroth" )
    }

    return( "zero" )
  }

  # Concatenate this a little later.
  if( n < 0 ) {
    result = "negative "
    n = abs( n )
  }

  # Do not spell out numbers greater than one hundred.
  if( n > 100 ) {
    # Comma-separated numbers.
    return( commas( n ) )
  }

  # Don't go beyond 100.
  if( n == 100 ) {
    if( ordinal ) {
      return( "one hundredth" )
    }

    return( "one hundred" )
  }

  # Samuel Langhorne Clemens noted English has too many exceptions.
  small = c(
    "one", "two", "three", "four", "five",
    "six", "seven", "eight", "nine", "ten",
    "eleven", "twelve", "thirteen", "fourteen", "fifteen",
    "sixteen", "seventeen", "eighteen", "nineteen"
  )

  ord_small = c(
    "first", "second", "third", "fourth", "fifth",
    "sixth", "seventh", "eighth", "ninth", "tenth",
    "eleventh", "twelfth", "thirteenth", "fourteenth", "fifteenth",
    "sixteenth", "seventeenth", "eighteenth", "nineteenth", "twentieth"
  )

  # After this, the number (n) is between 20 and 99.
  if( n < 20 ) {
    if( ordinal ) {
      return( .subset( ord_small, n %% 100 ) )
    }

    return( .subset( small, n %% 100 ) )
  }

  tens = c( "",
    "twenty", "thirty", "forty", "fifty",
    "sixty", "seventy", "eighty", "ninety"
  )

  ord_tens = c( "",
    "twentieth", "thirtieth", "fortieth", "fiftieth",
    "sixtieth", "seventieth", "eightieth", "ninetieth"
  )

  ones_index = n %% 10
  n = n %/% 10

  # No number in the ones column, so the number must be a multiple of ten.
  if( ones_index == 0 ) {
    if( ordinal ) {
      return( .subset( ord_tens, n ) )
    }

    return( .subset( tens, n ) )
  }

  # Find the value from the ones column.
  if( ordinal ) {
    unit_1 = .subset( ord_small, ones_index )
  }
  else {
    unit_1 = .subset( small, ones_index )
  }

  # Find the tens column.
  unit_10 = .subset( tens, n )

  # Hyphenate the tens and the ones together.
  paste0( unit_10, paste0( "-", unit_1 ) )
}

# -----------------------------------------------------------------------------
# Returns a number as a comma-delimited string. This is a work-around
# until Renjin fixes https://github.com/bedatadriven/renjin/issues/338
#
# @param n An integer value to delimit with commas
# -----------------------------------------------------------------------------
commas <- function( n ) {
  s <- sprintf( "%03.0f", n %% 1000 )
  n <- n %/% 1000

  while( n > 0 ) {
    s <- paste0( sprintf( "%03.0f", n %% 1000 ), ',', s )
    n <- n %/% 1000
  }

  gsub( '^0*', '', s )
}

