# ------------------------------------------------------------------------
# Copyright 2019 Dave Jarvis
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# ------------------------------------------------------------------------

# ------------------------------------------------------------------------
# The purpose of this software is to show temperature and CO2 measurements
# on the same graph.
# ------------------------------------------------------------------------
library( 'ggplot2' )
library( 'ggrepel' )
library( 'extrafont' )
library( 'scales' )
library( 'pander' )

# ------------------------------------------------------------------------
# Load, merge, and rescale climate temperature data in terms of CO2 PPM.
# The data sources must be in CSV format with the following headers:
#
# Year,Value
#
# The temperature and CO2 data files must be named:
#
# * global-temperature.csv
# * global-co2.csv
# ------------------------------------------------------------------------
load_climate_data <- function( year = 1880 ) {
  # Read and average the annual temperature data
  t_data <- read.csv( 'global-temperature.csv', header = TRUE )
  #t_data <- aggregate( t_data$Value, by=list(t_data$Year), FUN=mean )
  names( t_data ) <- c( "Year", "Value" )

  # Read the global CO2 PPM measurement data
  c_data <- read.csv( 'global-co2.csv', header = TRUE )

  merged = merge( t_data, c_data, by = "Year" )
  merged = subset( merged, Year >= year )
  names( merged ) <- c( "year", "temp", "ppm" )

  # Scale the temperature data relative to the CO2 measurements
  scaled <- rescale( merged$temp, to = range( merged$ppm ) )

  return( data.frame( merged, scaled ) )
}

# ------------------------------------------------------------------------
# Generate a climate graph using global temperature and CO2 data.
#
# The default font is Lato Heavy.
# ------------------------------------------------------------------------
graph_climate <- function( year = 1880 ) {
  climate <- load_climate_data( year )

  c_aes <- aes( x = year, y = ppm )
  t_aes <- aes( x = year, y = scaled )

  # Theme palette
  colour_black = '#020202'
  colour_gray = '#7f7f7f'
  colour_navy = '#022D51'
  colour_teal = '#065c5a'
  colour_orange = '#fc8f12'
  colour_white = '#fefefe'
  colour_yellow = '#ffe433'

  colour_temp = colour_orange
  colour_ppm = colour_yellow

  # Reusable elements
  f_family <- 'Lato Heavy'
  f_colour <- colour_navy
  f_size <- 26
  line_size <- 0.75

  # Temperature line
  t_line <- geom_line( t_aes,
    colour = colour_temp, size = line_size )

  # CO2 PPM area and line
  c_area <- stat_smooth( c_aes, method = 'loess', span = 0.1, geom = 'area',
    fill = colour_white, alpha = 0.3 )

  c_line <- geom_smooth( c_aes, method = 'loess', span = 0.1,
    colour = colour_ppm, size = line_size, se = FALSE )

  # Compute grid line extents
  y_coord_offset <- 20
  y_min <- trunc( min( climate$ppm ) / 100 ) * 100 + 50
  y_max <- round( max( climate$ppm ) / 100 ) * 100 + 20

  x_min <- min( climate$year )
  x_max <- max( climate$year )

  pad_sx <- seq( x_min, x_min )
  pad_sy <- seq( y_min, y_max, by = 20 )

  # Calculate scaling for secondary axis to align with primary y axis
  r_ppm <- range( climate$ppm )
  r_temp <- range( climate$temp )

  scale <- diff( r_ppm ) / diff( r_temp )
  temp_scale_breaks <- ((pad_sy - r_ppm[1]) / scale) + r_temp[1]

  # Compute axis scale transformation for temperature range
  temp_trans <- ~ ((. - r_ppm[1]) / scale) + r_temp[1]

  scale_x <- scale_x_continuous( expand = c( 0, 0 ), breaks = pad_sx )
  scale_y <- scale_y_continuous(
    expand = c( 0, 0 ),
    breaks = pad_sy,
    sec.axis = sec_axis(
      name = "°C",
      trans = temp_trans,
      breaks = temp_scale_breaks,
      # Hyphen and minus symbols differ typographically in width
      labels = gsub( "-", "−", sprintf( "%+.1f", temp_scale_breaks ) )
    )
  )

  # Set graph labels
  label_axis_y <- ylab( 'PPM' )

  # Custom location for the minimum year
  label_axis_x_min <- annotate(
    geom="text", -Inf, -Inf,
    label = x_min,
    vjust = 1.5,
    hjust = 0,
    family = f_family,
    colour = f_colour,
    size = f_size / 2.5,
  )

  # Custom location for the maximum year
  label_axis_x_max <- annotate(
    geom="text", Inf, -Inf,
    label = x_max,
    vjust = 1.5,
    hjust = 1,
    family = f_family,
    colour = f_colour,
    size = f_size / 2.5,
  )

  # Graph theme, no space around margin
  theme <- theme(
    plot.margin = margin( t = 1.8, b = 1.2, l = 0.2, r = 1.0, "cm" ),
    text = element_text(
      family = f_family,
      colour = f_colour,
    ),
    panel.background = element_rect( fill = f_colour ),
    panel.border = element_rect( colour = f_colour, fill = NA, size = 0.0 ),
    panel.grid.minor = element_blank(),
    panel.grid.major.x = element_line( colour = f_colour ),
    panel.grid.major.y = element_line(
      colour = colour_white,
      linetype = 'dotted',
    ),
    axis.ticks.length = unit( 1, "cm" ),
    axis.title.x = element_blank(),
    axis.title.y = element_text(
      size = f_size - 8,
      angle = 0,
      margin = margin( t = 0, r = -(f_size * 2) + 10, b = 0, l = 0 ),
      vjust = 1.05,
    ),
    axis.title.y.right = element_text(
      angle = 0,
      size = f_size - 8,
      color = f_colour,
      margin = margin( t = 0, r = 0, b = 0, l = -f_size - 18 ),
      vjust = 1.05,
    ),
    axis.text.x = element_blank(),
    axis.text.y = element_text(
      family = f_family,
      colour = f_colour,
      size = f_size,
    ),
    axis.ticks = element_blank(),
  )

  # Force label placement to be determinstic
  set.seed( 100 )

  # Common label styling
  label_font_size <- 5
  label_box_padding <- 4
  label_direction <- 'x'
  label_radius <- 0
  label_padding <- 0.5

  # Ensure the labels repel from all data points
  climate$ppm_label <- ""
  climate$temp_label <- ""

  # Place the CO2 label on the lower-half of the graph
  length_ppm = length( climate$ppm )
  half_ppm = tail( climate$ppm, length_ppm / 2 )
  ppm_median_index <-
    which.min( abs( half_ppm - median( half_ppm ) ) )
  climate$ppm_label[ ppm_median_index ] <- toupper( "Carbon Dioxide" )

  line_label_ppm <- geom_label_repel(
    c_aes,
    colour = f_colour,
    label = climate$ppm_label,
    label.r = label_radius,
    label.padding = label_padding,
    box.padding = label_box_padding,
    family = f_family,
    size = label_font_size,
    fill = colour_ppm,
    segment.colour = colour_ppm,
    vjust = -1,
    hjust = 1,
    na.rm = TRUE
  )

  # Place the temperature label on upper-half of the graph
  length_temp = length( climate$scaled )
  half_temp = head( climate$scaled, length_temp / 2 )
  temp_median_index <-
    which.min( abs( half_temp - median( half_temp ) ) ) + (length_temp / 2)
  climate$temp_label[ temp_median_index ] <- toupper( "Temperature" )

  line_label_temp <- geom_label_repel(
    t_aes,
    colour = f_colour,
    label = climate$temp_label,
    label.r = label_radius,
    label.padding = label_padding,
    box.padding = label_box_padding,
    family = f_family,
    size = label_font_size,
    fill = colour_temp,
    segment.colour = colour_temp,
    vjust = -4,
    hjust = 1,
    na.rm = TRUE
  )

  # Disable clipping and crop the graph to sensible limits
  coord <- coord_cartesian(
    ylim = c( y_min + y_coord_offset, y_max ),
    clip = 'off',
  )

  # Generate plot using theme, axis, and label customisations
  plot <- ggplot( climate, xaxs = 'i', yaxs = 'i' )
  plot <- plot + theme
  plot <- plot + t_line + c_area + c_line
  plot <- plot + scale_x + scale_y + coord
  plot <- plot + line_label_ppm + line_label_temp
  plot <- plot + label_axis_y
  plot <- plot + label_axis_x_min + label_axis_x_max

  return( plot )
}

# ------------------------------------------------------------------------
# Call pander's evals to store write the image into a temporary file that
# can be included in an R Markdown document.
#
# @return The path to the generated image.
# ------------------------------------------------------------------------
plot_graph_climate <- function() {
  evals(
    'graph_climate()',
    graph.output = "svg"
  )[[1]]$result[1]
}

