# -----------------------------------------------------------------------------
# Copyright 2019 Dave Jarvis
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# Rounds to the nearest base value (e.g., round to nearest 10).
#
# @param base The nearest value to round to.
# -----------------------------------------------------------------------------
round.up <- function( n, base = 5 ) {
  base * round( n / base )
}

# -----------------------------------------------------------------------------
# Calculates the amount of time to travel between two locations at a given
# velocity, in minutes.
#
# @param lat1 The source latitude.
# @param lon1 The source longitude.
# @param lat2 The destination latitude.
# @param lon2 The destination longitude.
# @param velocity Given in km/h.
#
# @return The travel time between two locations, in minutes.
# -----------------------------------------------------------------------------
travel.time <- function( lat1, lon1, lat2, lon2, velocity ) {
  km = distance( lat1, lon1, lat2, lon2 ) / 1000

  km / velocity * 60
}

# -----------------------------------------------------------------------------
# Computes the distance between two locations in meters. This uses an online
# map API and therefore an Internet connection is required for an accurate
# result. If no connection is found, this will use the Haversine formula
# to provide a rough estimate for the distance.
#
# @param lat1/lon1 The source latitude and longitude.
# @param lat2/lon2 The destination latitude and longitude.
#
# @return Distance between the two coordinates via roadways.
# -----------------------------------------------------------------------------
distance <- function( lat1, lon1, lat2, lon2 ) {
  token <- readLines( "token.txt" )

  # Create the URL to use to get the distance data.
  api <- paste0(
    'https://api.mapbox.com/directions/v5/mapbox/driving/',
    lon1,
    ',',
    lat1,
    ';',
    lon2,
    ',',
    lat2,
    paste0( '?access_token=', token )
  )

  tryCatch({
    # Download the document with distance information.
    doc <- fromJSON( file=api )

    # Obtain the distance in meters.
    meters = doc[1]$routes[[1]]$distance

    # Return the distance.
    as.numeric( meters )
  },
  warning = function( w ) {
    haversine( lat1, lon1, lat2, lon2 )
  },
  error = function( e ) {
    haversine( lat1, lon1, lat2, lon2 )
  })
}

# -----------------------------------------------------------------------------
# Computes linear distance between two points using Haversine formula.
# Although Earth is an oblate spheroid, this will produce results close
# enough for most purposes.
#
# @param lat1 The source latitude.
# @param lon1 The source longitude.
# @param lat2 The destination latitude.
# @param lon2 The destination longitude.
# @param radius The radius of the sphere.
#
# @return The distance between the two coordinates in meters.
# -----------------------------------------------------------------------------
haversine <- function( lat1, lon1, lat2, lon2, radius = 6371 ) {
  # Convert decimal degrees to radians
  lon1 = lon1 * pi / 180
  lon2 = lon2 * pi / 180
  lat1 = lat1 * pi / 180
  lat2 = lat2 * pi / 180

  # Haversine formula
  dlon = lon2 - lon1
  dlat = lat2 - lat1
  a = sin( dlat / 2 ) ** 2 + cos( lat1 ) * cos( lat2 ) * sin( dlon / 2 ) ** 2
  c = 2 * atan2( sqrt( a ), sqrt( 1-a ) )

  return( radius * c * 1000 )
}

